# -*- coding: utf-8 -*-
# Copyright (C) 2020 Adrien Delle Cave
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.classes.abstract.db"""

import abc
import logging
import os

from dwho.classes.errors import DWhoConfigurationError
from sonicprobe.libs import anysql, mysql_config_parser, urisup

LOG           = logging.getLogger('xtrargon_ng.classes.abstract.db')

DB_URI_PARAMS = (('autocommit', '1'),
                 ('charset', 'utf8'),
                 ('cursorclass', 'Dict'),
                 ('wait_timeout', '2147483'),
                 ('connect_timeout', '10'))

DB_DISTRIB_MARIADB = 'mariadb'
DB_DISTRIB_PERCONA = 'percona'
DB_DISTRIB_ORACLE  = 'oracle'


class XtrargonDBBase(object): # pylint: disable-msg=useless-object-inheritance
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.dbconf          = None
        self.dburi           = None

        self._dbconn         = None
        self._dbcursor       = None
        self._sql_thread_run = False

    def _build_mysql_uri(self, host = None, port = None):
        host_cfg = mysql_config_parser.MYSQL_DEFAULT_HOST
        port_cfg = mysql_config_parser.MYSQL_DEFAULT_PORT

        if self.dbconf.has_option('client', 'host'):
            host_cfg = self.dbconf.get('client', 'host')

        if self.dbconf.has_option('client', 'port'):
            port_cfg = self.dbconf.get('client', 'port')

        if host:
            host_cfg = host

        if port:
            port_cfg = port

        uri = ('mysql',
               (self.dbconf.get('client', 'user'),
                self.dbconf.get('client', 'password'),
                host_cfg,
                port_cfg),
               None,
               DB_URI_PARAMS,
               None)

        return urisup.uri_help_unsplit(uri)

    @staticmethod
    def _check_db_server(uri):
        dbconn = None
        try:
            dbconn = anysql.connect_by_uri(uri)
            if not dbconn.is_connected():
                raise DWhoConfigurationError("unable to connect to database")
        finally:
            if dbconn:
                dbconn.close()

    # pylint: disable-msg=attribute-defined-outside-init
    def init_db(self, module = None, default_cnf_file = 'client.cnf', host = None, port = None):
        if module is None:
            module = self

        self.dbconf = mysql_config_parser.MySQLConfigVersion(
            default_file = default_cnf_file,
            custom_file  = os.path.join('client.d', "%s.cnf" % module.server_id),
            config_dir   = module.config.get('_config_directory')).get_client(check_version = False)

        self.dburi  = self._build_mysql_uri(host, port)

    def create_connection(self):
        try:
            self._check_connection()
        except Exception:
            self.close_connection()
            self._dbconn   = anysql.connect_by_uri(self.dburi)
            self._dbcursor = self._dbconn.cursor()

    def _check_connection(self):
        if not self._dbconn:
            raise LookupError("dbconn uninitialized")

        if not self._dbconn.is_connected():
            raise LookupError("dbconn unconnected")

        if not self._dbcursor:
            raise LookupError("dbcursor uninitialized")

    def _get_sql_variable(self, varname):
        self._dbcursor.query("SHOW VARIABLES LIKE '%s'" % varname)
        row = self._dbcursor.fetchone(raw = True)
        if row:
            return row['Value']

        return None

    def which_distrib(self):
        version_comment = self._get_sql_variable('version_comment').lower()
        if version_comment:
            version_comment = version_comment.lower()
            if DB_DISTRIB_MARIADB in version_comment:
                return DB_DISTRIB_MARIADB

            if DB_DISTRIB_PERCONA in version_comment:
                return DB_DISTRIB_PERCONA

        if self._get_sql_variable('innodb_fast_checksum') \
           or self._get_sql_variable('have_backup_locks'):
            return DB_DISTRIB_PERCONA

        return DB_DISTRIB_ORACLE

    def get_slave_status(self):
        self._dbcursor.query("SHOW SLAVE STATUS")
        row = self._dbcursor.fetchone(raw = True)
        if row:
            return row

        return None

    def test_connection(self):
        if getattr(self, '_killed', False):
            raise GeneratorExit()

        self._check_connection()

        self._dbcursor.query("SELECT 1",
                             None,
                             None)
        row = self._dbcursor.fetchone()
        if row:
            return int(row[0]) > 0

        return False

    def check_is_master(self):
        if getattr(self, '_killed', False):
            raise GeneratorExit()

        self._check_connection()

        server_id = self._get_sql_variable('server_id')
        if not server_id or int(server_id) == 0:
            return False

        self._dbcursor.query("SELECT 1 FROM information_schema.processlist "
                             " WHERE command = ?"
                             " LIMIT 1",
                             None,
                             ('Binlog Dump',))
        row = self._dbcursor.fetchone()
        if row:
            return int(row[0]) > 0

        return False

    def check_is_slave(self):
        if getattr(self, '_killed', False):
            raise GeneratorExit()

        self._check_connection()

        server_id = self._get_sql_variable('server_id')
        if not server_id or int(server_id) == 0:
            return False

        slave_status = self.get_slave_status()
        if not slave_status \
           or not slave_status['Read_Master_Log_Pos'] \
           or not slave_status['Slave_SQL_Running']:
            return False

        self._sql_thread_run = slave_status['Slave_SQL_Running'] == 'Yes'

        return slave_status['Slave_SQL_Running']

    def replica_io_running(self):
        slave_status = self.get_slave_status()
        if not slave_status \
           or not slave_status['Slave_IO_Running']:
            return None

        return slave_status['Slave_IO_Running'] == 'Yes'

    def close_connection(self):
        if self._dbcursor:
            try:
                self._dbcursor.close()
                self._dbcursor = None
            except Exception:
                pass

        if self._dbconn:
            try:
                self._dbconn.close()
                self._dbconn = None
            except Exception:
                pass
